"""
eobox.sampledata subpackage

The subpackage which holds data and functions to get file paths to data. 
sampledata is used in the examples and tests of other eobox sub-packages. 

For more information on the package content, visit [readthedocs](https://eo-box.readthedocs.raster/en/latest/eobox.sampledata.html).
"""

from . import base

from .base import get_dataset
